package flm.b4a.scrollview2d;

import java.util.List;

import android.content.Context;
import android.graphics.Rect;
import android.util.AttributeSet;
import android.view.FocusFinder;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.VelocityTracker;
import android.view.View;
import android.view.ViewConfiguration;
import android.view.ViewGroup;
import android.view.ViewParent;
import android.view.animation.AnimationUtils;
import android.widget.FrameLayout;
import android.widget.Scroller;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.objects.PanelWrapper;

/**
 * Layout container for a view hierarchy that can be scrolled by the user,
 * allowing it to be larger than the physical display.  A ScrollView2D
 * is a {@link FrameLayout}, meaning you should place one child in it
 * containing the entire contents to scroll; this child may itself be a layout
 * manager with a complex hierarchy of objects.
 */
@BA.Hide
public class TwoDScrollView extends FrameLayout
{
	PanelWrapper pw;

	private static final int ANIMATED_SCROLL_GAP = 250;

	private static final float MAX_SCROLL_FACTOR = 0.5f;

	private long mLastScroll;

	private final Rect mTempRect = new Rect();
	private Scroller mScroller;

	protected boolean DisallowInterceptTouchEvent = true; 

	/**
	 * Position of the last motion event.
	 */
	private int mLastMotionX;
	private int mLastMotionY;

	/**
	 * Position du scroll avant un dplacement du doigt
	 */
	private int mStartX;
	private int mStartY;

	/**
	 * True when the layout has changed but the traversal has not come through yet.
	 * Ideally the view hierarchy would keep track of this for us.
	 */
	private boolean mIsLayoutDirty = true;

	/**
	 * The child to give focus to in the event that a child has requested focus while the
	 * layout is dirty. This prevents the scroll from being wrong if the child has not been
	 * laid out before requesting focus.
	 */
	private View mChildToScrollTo = null;

	/**
	 * True if the user is currently dragging this ScrollView around. This is
	 * not the same as 'is being flinged', which can be checked by
	 * mScroller.isFinished() (flinging begins when the user lifts his finger).
	 */
	private boolean mIsBeingDragged = false;

	/**
	 * Determines speed during touch scrolling
	 */
	private VelocityTracker mVelocityTracker;

	/**
	 * When set to true, the scroll view measure its child to make it fill the currently
	 * visible area.
	 */
	private boolean mFillViewport;

	/**
	 * Whether arrow scrolling is animated.
	 */
	private boolean mSmoothScrollingEnabled = true;

	private int mTouchSlop;
	private int mMinimumVelocity;
	private int mMaximumVelocity;

	/**
	 * ID of the active pointer. This is used to retain consistency during
	 * drags/flings if multiple pointers are used.
	 */
	private int mActivePointerId = INVALID_POINTER;

	/**
	 * Sentinel value for no current active pointer.
	 * Used by {@link #mActivePointerId}.
	 */
	private static final int INVALID_POINTER = -1;

	public TwoDScrollView(Context context) {
		this(context, null);
	}

	public TwoDScrollView(Context context, AttributeSet attrs) {
		this(context, attrs, android.R.attr.scrollViewStyle);
	}

	public TwoDScrollView(Context context, AttributeSet attrs, int defStyle) {
		super(context, attrs, defStyle);
		initTwoDScrollView();
	}

	private void initTwoDScrollView() {
		mScroller = new Scroller(getContext());
		setFocusable(true);
		setDescendantFocusability(FOCUS_AFTER_DESCENDANTS);
		setWillNotDraw(false);
		final ViewConfiguration configuration = ViewConfiguration.get(getContext());
		mTouchSlop = configuration.getScaledTouchSlop();
		mMinimumVelocity = configuration.getScaledMinimumFlingVelocity();
		mMaximumVelocity = configuration.getScaledMaximumFlingVelocity();
		setHorizontalFadingEdgeEnabled(true);
		setVerticalFadingEdgeEnabled(true);
		setHorizontalScrollBarEnabled(true);
		setVerticalScrollBarEnabled(true);
	}

	@Override
	protected float getTopFadingEdgeStrength() {
		if (getChildCount() == 0) {
			return 0.0f;
		}

		final int length = getVerticalFadingEdgeLength();
		if (getScrollY() < length) {
			return getScrollY() / (float) length;
		}

		return 1.0f;
	}

	@Override
	protected float getBottomFadingEdgeStrength() {
		if (getChildCount() == 0) {
			return 0.0f;
		}

		final int length = getVerticalFadingEdgeLength();
		final int bottomEdge = getHeight() - getPaddingBottom();
		final int span = getChildAt(0).getBottom() - getScrollY() - bottomEdge;
		if (span < length) {
			return span / (float) length;
		}

		return 1.0f;
	}

	@Override
	protected float getLeftFadingEdgeStrength() {
		if (getChildCount() == 0) {
			return 0.0f;
		}

		final int length = getHorizontalFadingEdgeLength();
		if (getScrollX() < length) {
			return getScrollX() / (float) length;
		}

		return 1.0f;
	}

	@Override
	protected float getRightFadingEdgeStrength() {
		if (getChildCount() == 0) {
			return 0.0f;
		}

		final int length = getHorizontalFadingEdgeLength();
		final int rightEdge = getWidth() - getPaddingRight();
		final int span = getChildAt(0).getRight() - getScrollX() - rightEdge;
		if (span < length) {
			return span / (float) length;
		}

		return 1.0f;
	}

	/**
	 * @return The maximum amount this scroll view will scroll in response to
	 *   an arrow event.
	 */
	public int getMaxScrollAmountV() {
		return (int) (MAX_SCROLL_FACTOR * getHeight());
	}
	public int getMaxScrollAmountH() {
		return (int) (MAX_SCROLL_FACTOR * getWidth());
	}

	@Override
	public void addView(View child) {
		if (getChildCount() > 0) {
			throw new IllegalStateException("ScrollView2D can host only one direct child");
		}

		super.addView(child);
	}

	@Override
	public void addView(View child, int index) {
		if (getChildCount() > 0) {
			throw new IllegalStateException("ScrollView2D can host only one direct child");
		}

		super.addView(child, index);
	}

	@Override
	public void addView(View child, ViewGroup.LayoutParams params) {
		if (getChildCount() > 0) {
			throw new IllegalStateException("ScrollView2D can host only one direct child");
		}

		super.addView(child, params);
	}

	@Override
	public void addView(View child, int index, ViewGroup.LayoutParams params) {
		if (getChildCount() > 0) {
			throw new IllegalStateException("ScrollView2D can host only one direct child");
		}

		super.addView(child, index, params);
	}

	/**
	 * @return Returns true this ScrollView can be scrolled
	 */
	private boolean canScrollV() {
		View child = getChildAt(0);
		if (child != null) {
			int childHeight = child.getHeight();
			return getHeight() < childHeight + getPaddingTop() + getPaddingBottom();
		}
		return false;
	}
	private boolean canScrollH() {
		View child = getChildAt(0);
		if (child != null) {
			int childWidth = child.getWidth();
			return getWidth() < childWidth + getPaddingLeft() + getPaddingRight() ;
		}
		return false;
	}

	/**
	 * Indicates whether this ScrollView's content is stretched to fill the viewport.
	 *
	 * @return True if the content fills the viewport, false otherwise.
	 */
	public boolean isFillViewport() {
		return mFillViewport;
	}

	/**
	 * Indicates this ScrollView whether it should stretch its content height to fill
	 * the viewport or not.
	 *
	 * @param fillViewport True to stretch the content's height to the viewport's
	 *        boundaries, false otherwise.
	 */
	public void setFillViewport(boolean fillViewport) {
		if (fillViewport != mFillViewport) {
			mFillViewport = fillViewport;
			requestLayout();
		}
	}

	/**
	 * @return Whether arrow scrolling will animate its transition.
	 */
	public boolean isSmoothScrollingEnabled() {
		return mSmoothScrollingEnabled;
	}

	/**
	 * Set whether arrow scrolling will animate its transition.
	 * @param smoothScrollingEnabled whether arrow scrolling will animate its transition
	 */
	public void setSmoothScrollingEnabled(boolean smoothScrollingEnabled) {
		mSmoothScrollingEnabled = smoothScrollingEnabled;
	}

	@Override
	protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
		super.onMeasure(widthMeasureSpec, heightMeasureSpec);

		if (!mFillViewport) {
			return;
		}

		final int heightMode = MeasureSpec.getMode(heightMeasureSpec);
		final int widthMode = MeasureSpec.getMode(widthMeasureSpec);
		if (heightMode == MeasureSpec.UNSPECIFIED && widthMode == MeasureSpec.UNSPECIFIED) {
			return;
		}

		if (getChildCount() > 0) {
			final View child = getChildAt(0);
			final FrameLayout.LayoutParams lp = (LayoutParams) child.getLayoutParams();
			int height = getMeasuredHeight();
			int width = getMeasuredWidth();
			if (child.getMeasuredHeight() < height && heightMode != MeasureSpec.UNSPECIFIED) {

				int childWidthMeasureSpec = getChildMeasureSpec(widthMeasureSpec, getPaddingLeft() + getPaddingRight(), lp.width);
				height -= getPaddingTop();
				height -= getPaddingBottom();
				int childHeightMeasureSpec = MeasureSpec.makeMeasureSpec(height, MeasureSpec.EXACTLY);

				child.measure(childWidthMeasureSpec, childHeightMeasureSpec);
			}
			if (child.getMeasuredWidth() < width && widthMode != MeasureSpec.UNSPECIFIED) {

				int childHeightMeasureSpec = getChildMeasureSpec(heightMeasureSpec, getPaddingTop() + getPaddingBottom(), lp.height);
				width -= getPaddingLeft();
				width -= getPaddingRight();
				int childWidthMeasureSpec = MeasureSpec.makeMeasureSpec(width, MeasureSpec.EXACTLY);

				child.measure(childWidthMeasureSpec, childHeightMeasureSpec);
			}
		}
	}

	@Override
	public boolean dispatchKeyEvent(KeyEvent event) {
		// Let the focused view and/or our descendants get the key first
		return super.dispatchKeyEvent(event) || executeKeyEvent(event);
	}

	/**
	 * You can call this function yourself to have the scroll view perform
	 * scrolling from a key event, just as if the event had been dispatched to
	 * it by the view hierarchy.
	 *
	 * @param event The key event to execute.
	 * @return Return true if the event was handled, else false.
	 */

	private boolean CannotScrollV() {
		if (isFocused()) {
			View currentFocused = findFocus();
			if (currentFocused == this) currentFocused = null;
			View nextFocused = FocusFinder.getInstance().findNextFocus(this,
					currentFocused, View.FOCUS_DOWN);
			return nextFocused != null
					&& nextFocused != this
					&& nextFocused.requestFocus(View.FOCUS_DOWN);
		}
		return false;
	}
	private boolean CannotScrollH() {
		if (isFocused()) {
			View currentFocused = findFocus();
			if (currentFocused == this) currentFocused = null;
			View nextFocused = FocusFinder.getInstance().findNextFocus(this,
					currentFocused, View.FOCUS_RIGHT);
			return nextFocused != null && nextFocused != this &&
					nextFocused.requestFocus(View.FOCUS_RIGHT);
		}
		return false;
	}

	public boolean executeKeyEvent(KeyEvent event) {
		mTempRect.setEmpty();

		boolean handled = false;
		if (event.getAction() == KeyEvent.ACTION_DOWN) {
			switch (event.getKeyCode()) {
			case KeyEvent.KEYCODE_DPAD_UP:
				if (!canScrollV()) return CannotScrollV(); 
				if (!event.isAltPressed()) {
					handled = arrowScrollV(View.FOCUS_UP);
				} else {
					handled = fullScroll(View.FOCUS_UP);
				}
				break;
			case KeyEvent.KEYCODE_DPAD_DOWN:
				if (!canScrollV()) return CannotScrollV(); 
				if (!event.isAltPressed()) {
					handled = arrowScrollV(View.FOCUS_DOWN);
				} else {
					handled = fullScroll(View.FOCUS_DOWN);
				}
				break;
			case KeyEvent.KEYCODE_DPAD_LEFT:
				if (!canScrollH()) return CannotScrollH(); 
				if (!event.isAltPressed()) {
					handled = arrowScrollH(View.FOCUS_LEFT);
				} else {
					handled = fullScroll(View.FOCUS_LEFT);
				}
				break;
			case KeyEvent.KEYCODE_DPAD_RIGHT:
				if (!canScrollH()) return CannotScrollH(); 
				if (!event.isAltPressed()) {
					handled = arrowScrollH(View.FOCUS_RIGHT);
				} else {
					handled = fullScroll(View.FOCUS_RIGHT);
				}
				break;
			case KeyEvent.KEYCODE_SPACE:
				if (!canScrollV()) return CannotScrollV(); 
				pageScroll(event.isShiftPressed() ? View.FOCUS_UP : View.FOCUS_DOWN);
				break;
			}
		}

		return handled;
	}

	private boolean inChild(int x, int y) {
		if (getChildCount() > 0) {
			final int scrollX = getScrollX();
			final int scrollY = getScrollY();
			final View child = getChildAt(0);
			return !(y < child.getTop() - scrollY
					|| y >= child.getBottom() - scrollY
					|| x < child.getLeft() - scrollX
					|| x >= child.getRight() - scrollX);
		}
		return false;
	}

	private void initOrResetVelocityTracker() {
		if (mVelocityTracker == null) {
			mVelocityTracker = VelocityTracker.obtain();
		} else {
			mVelocityTracker.clear();
		}
	}

	private void initVelocityTrackerIfNotExists() {
		if (mVelocityTracker == null) {
			mVelocityTracker = VelocityTracker.obtain();
		}
	}

	private void recycleVelocityTracker() {
		if (mVelocityTracker != null) {
			mVelocityTracker.recycle();
			mVelocityTracker = null;
		}
	}

	@Override
	public void requestDisallowInterceptTouchEvent(boolean disallowIntercept) {
		if (disallowIntercept) {
			recycleVelocityTracker();
		}
		super.requestDisallowInterceptTouchEvent(disallowIntercept);
	}

	@Override
	public boolean onInterceptTouchEvent(MotionEvent ev) {
		/*
		 * This method JUST determines whether we want to intercept the motion.
		 * If we return true, onMotionEvent will be called and we do the actual
		 * scrolling there.
		 */

		/*
		 * Shortcut the most recurring case: the user is in the dragging
		 * state and he is moving his finger.  We want to intercept this
		 * motion.
		 */
		final int action = ev.getAction();
		if ((action == MotionEvent.ACTION_MOVE) && (mIsBeingDragged)) {
			return true;
		}

		switch (action & MotionEvent.ACTION_MASK) {
		case MotionEvent.ACTION_MOVE: {
			/*
			 * mIsBeingDragged == false, otherwise the shortcut would have caught it. Check
			 * whether the user has moved far enough from his original down touch.
			 */

			/*
			 * Locally do absolute value.
			 * mLastMotionX and mLastMotionY are set to the x and y values of the down event.
			 */
			final int activePointerId = mActivePointerId;
			if (activePointerId == INVALID_POINTER) {
				// If we don't have a valid id, the touch down wasn't on content.
				break;
			}

			final int pointerIndex = ev.findPointerIndex(activePointerId);
			final int x = (int) ev.getX(pointerIndex);
			final int y = (int) ev.getY(pointerIndex);
			final int xDiff = Math.abs(x - mLastMotionX);
			final int yDiff = Math.abs(y - mLastMotionY);
			if (xDiff > mTouchSlop || yDiff > mTouchSlop) {
				mIsBeingDragged = true;
				mLastMotionX = x;
				mLastMotionY = y;
				initVelocityTrackerIfNotExists();
				mVelocityTracker.addMovement(ev);
				final ViewParent parent = getParent();
				if (parent != null) {
					parent.requestDisallowInterceptTouchEvent(DisallowInterceptTouchEvent);
				}
			}
			break;
		}

		case MotionEvent.ACTION_DOWN: {
			final int x = (int) ev.getX();
			final int y = (int) ev.getY();
			if (!inChild(x, y)) {
				mIsBeingDragged = false;
				recycleVelocityTracker();
				break;
			}

			/*
			 * Remember location of down touch.
			 * ACTION_DOWN always refers to pointer index 0.
			 */
			mLastMotionX = x;
			mLastMotionY = y;
			mStartX = getScrollX();
			mStartY = getScrollY();
			mActivePointerId = ev.getPointerId(0);

			initOrResetVelocityTracker();
			mVelocityTracker.addMovement(ev);

			/*
			 * If being flinged and user touches the screen, initiate drag;
			 * otherwise don't.  mScroller.isFinished should be false when
			 * being flinged.
			 */
			mIsBeingDragged = !mScroller.isFinished();
			break;
		}

		case MotionEvent.ACTION_CANCEL:
		case MotionEvent.ACTION_UP:
			/* Release the drag */
			mIsBeingDragged = false;
			mActivePointerId = INVALID_POINTER;
			recycleVelocityTracker();
			break;
		case MotionEvent.ACTION_POINTER_UP:
			onSecondaryPointerUp(ev);
			break;
		}

		/*
		 * The only time we want to intercept motion events is if we are in the
		 * drag mode.
		 */
		return mIsBeingDragged;
	}

	private void endDrag() {
		mIsBeingDragged = false;
		recycleVelocityTracker();
	}

	@Override
	public boolean onTouchEvent(MotionEvent ev) {
		initVelocityTrackerIfNotExists();
		mVelocityTracker.addMovement(ev);

		final int action = ev.getAction();

		switch (action & MotionEvent.ACTION_MASK) {
		case MotionEvent.ACTION_DOWN: {
			if (getChildCount() == 0) {
				return false;
			}
			if ((mIsBeingDragged = !mScroller.isFinished())) {
				final ViewParent parent = getParent();
				if (parent != null) {
					parent.requestDisallowInterceptTouchEvent(DisallowInterceptTouchEvent);
				}
			}

			/*
			 * If being flinged and user touches, stop the fling. isFinished
			 * will be false if being flinged.
			 */
			if (!mScroller.isFinished()) {
				mScroller.abortAnimation();
			}

			// Remember where the motion event started
			mLastMotionX = (int) ev.getX();
			mLastMotionY = (int) ev.getY();
			mStartX = getScrollX();
			mStartY = getScrollY();
			mActivePointerId = ev.getPointerId(0);
			break;
		}

		case MotionEvent.ACTION_MOVE: {
			final int activePointerIndex = ev.findPointerIndex(mActivePointerId);
			final int x = (int) ev.getX(activePointerIndex);
			final int y = (int) ev.getY(activePointerIndex);
			int deltaX = mLastMotionX - x;
			int deltaY = mLastMotionY - y;
			if (!mIsBeingDragged && (Math.abs(deltaX) > mTouchSlop || Math.abs(deltaY) > mTouchSlop)) {
				final ViewParent parent = getParent();
				if (parent != null) {
					parent.requestDisallowInterceptTouchEvent(DisallowInterceptTouchEvent);
				}
				mIsBeingDragged = true;
				if (Math.abs(deltaX) > mTouchSlop) {
					if (deltaX > 0) {
						deltaX -= mTouchSlop;
					} else {
						deltaX += mTouchSlop;
					}
				}
				if (Math.abs(deltaY) > mTouchSlop) {
					if (deltaY > 0) {
						deltaY -= mTouchSlop;
					} else {
						deltaY += mTouchSlop;
					}
				}
			}
			if (mIsBeingDragged) {
				// Scroll to follow the motion event
				mLastMotionX = x;
				mLastMotionY = y;
				scrollBy(deltaX, deltaY);
			}
			break;
		}
		case MotionEvent.ACTION_UP:
			if (mIsBeingDragged) {
				final VelocityTracker velocityTracker = mVelocityTracker;
				velocityTracker.computeCurrentVelocity(1000, mMaximumVelocity);
				int initialVelocityH = (int) velocityTracker.getXVelocity();
				int initialVelocityV = (int) velocityTracker.getYVelocity();

				if (getChildCount() > 0) {
					if (Math.abs(initialVelocityH) > mMinimumVelocity || Math.abs(initialVelocityV) > mMinimumVelocity) {
						fling(-initialVelocityH, -initialVelocityV);
					}
					else {
						final int deltaX = getScrollX() - mStartX;
						final int deltaY = getScrollY() - mStartY;
						FindNewFocusedView(deltaX, deltaY);
					}
				}

				mActivePointerId = INVALID_POINTER;
				endDrag();
			}
			break;
		case MotionEvent.ACTION_CANCEL:
			if (mIsBeingDragged && getChildCount() > 0) {
				mActivePointerId = INVALID_POINTER;
				endDrag();
			}
			break;
		case MotionEvent.ACTION_POINTER_UP:
			onSecondaryPointerUp(ev);
			break;
		}
		return true;
	}

	private void onSecondaryPointerUp(MotionEvent ev) {
		final int pointerIndex = (ev.getAction() & MotionEvent.ACTION_POINTER_ID_MASK) >>
		MotionEvent.ACTION_POINTER_ID_SHIFT;
		final int pointerId = ev.getPointerId(pointerIndex);
		if (pointerId == mActivePointerId) {
			// This was our active pointer going up. Choose a new
			// active pointer and adjust accordingly.
			final int newPointerIndex = pointerIndex == 0 ? 1 : 0;
			mLastMotionX = (int) ev.getX(newPointerIndex);
			mLastMotionY = (int) ev.getY(newPointerIndex);
			mActivePointerId = ev.getPointerId(newPointerIndex);
			if (mVelocityTracker != null) {
				mVelocityTracker.clear();
			}
		}
	}

	public void FindNewFocusedView(final int LeftFocus, final int TopFocus) {
		/* LeftFocus/TopFocus:
		 * 		< 0: le plus  droite/le plus en bas
		 * 		= 0: aucune prfrence
		 * 		> 0: le plus  gauche/le plus en haut
		 */
		View newFocused;
		if (mScroller.isFinished()) {
			newFocused = findFocusableViewInMyBounds(TopFocus, getScrollY(), LeftFocus, getScrollX(), findFocus());
		}
		else {
			newFocused = findFocusableViewInMyBounds(TopFocus, mScroller.getFinalY(), LeftFocus, mScroller.getFinalX(), findFocus());
		}
		if (newFocused == null) {
			newFocused = this;
		}
		if (newFocused != findFocus()) {
			newFocused.requestFocus();
		}
	}

	/**
	 * <p>
	 * Finds the next focusable component that fits in this View's bounds
	 * (excluding fading edges) pretending that this View's top is located at
	 * the parameter top.
	 * </p>
	 *
	 * @param topFocus           look for a candidate is the one at the top of the bounds
	 *                           if topFocus is true, or at the bottom of the bounds if topFocus is
	 *                           false
	 * @param top                the top offset of the bounds in which a focusable must be
	 *                           found (the fading edge is assumed to start at this position)
	 * @param preferredFocusable the View that has highest priority and will be
	 *                           returned if it is within my bounds (null is valid)
	 * @return the next focusable component in the bounds or null if none can be
	 *         found
	 */
	private View findFocusableViewInMyBounds(final int topFocus, final int top, 
			final int leftFocus, final int left,
			View preferredFocusable)
	{
		/*
		 * The fading edge's transparent side should be considered for focus
		 * since it's mostly visible, so we divide the actual fading edge length
		 * by 2.
		 */
		final int fadingEdgeLengthV = getVerticalFadingEdgeLength() / 2;
		final int topWithoutFadingEdge = top + fadingEdgeLengthV;
		final int bottomWithoutFadingEdge = top + getHeight() - fadingEdgeLengthV;

		final int fadingEdgeLengthH = getHorizontalFadingEdgeLength() / 2;
		final int leftWithoutFadingEdge = left + fadingEdgeLengthH;
		final int rightWithoutFadingEdge = left + getWidth() - fadingEdgeLengthH;

		if (preferredFocusable != null) {
			preferredFocusable.getDrawingRect(mTempRect);
			offsetDescendantRectToMyCoords(preferredFocusable, mTempRect);
			if (mTempRect.top < bottomWithoutFadingEdge && mTempRect.bottom > topWithoutFadingEdge &&
					mTempRect.left < rightWithoutFadingEdge && mTempRect.right > leftWithoutFadingEdge) {
				return preferredFocusable;
			}
		}

		return findFocusableViewInBounds(topFocus, topWithoutFadingEdge, bottomWithoutFadingEdge,
				leftFocus, leftWithoutFadingEdge, rightWithoutFadingEdge);
	}

	/**
	 * <p>
	 * Finds the next focusable component that fits in the specified bounds.
	 * </p>
	 *
	 * @param topFocus look for a candidate is the one at the top of the bounds
	 *                 if topFocus is true, or at the bottom of the bounds if topFocus is
	 *                 false
	 * @param top      the top offset of the bounds in which a focusable must be
	 *                 found
	 * @param bottom   the bottom offset of the bounds in which a focusable must
	 *                 be found
	 * @return the next focusable component in the bounds or null if none can
	 *         be found
	 */
	private View findFocusableViewInBounds(final int topFocus, int top, int bottom,
			final int leftFocus, int left, int right) {
		/* LeftFocus/TopFocus:
		 * 		< 0: le plus  droite/le plus en bas
		 * 		= 0: aucune prfrence
		 * 		> 0: le plus  gauche/le plus en haut
		 */
		List<View> focusables = getFocusables(View.FOCUS_FORWARD);
		View focusCandidate = null;
		Rect candidateRect = new Rect();

		/*
		 * A fully contained focusable is one where its top is below the bound's
		 * top, and its bottom is above the bound's bottom. A partially
		 * contained focusable is one where some part of it is within the
		 * bounds, but it also has some part that is not within bounds.  A fully contained
		 * focusable is preferred to a partially contained focusable.
		 */
		boolean foundFullyContainedFocusable = false;

		int count = focusables.size();
		for (int i = 0; i < count; i++) {
			View view = focusables.get(i);
			view.getDrawingRect(mTempRect);
			offsetDescendantRectToMyCoords(view, mTempRect);

			if (top < mTempRect.bottom && mTempRect.top < bottom && left < mTempRect.right && mTempRect.left < right) {
				/*
				 * the focusable is in the target area, it is a candidate for
				 * focusing
				 */

				final boolean viewIsFullyContained = (top < mTempRect.top) && (mTempRect.bottom < bottom)
						&& (left < mTempRect.left) && (mTempRect.right < right);

				if (focusCandidate == null) {
					/* No candidate, take this one */
					focusCandidate = view;
					candidateRect.set(mTempRect);
					foundFullyContainedFocusable = viewIsFullyContained;
				}
				else {
					boolean viewIsCloserToBoundaryV = true;
					if (topFocus != 0) {
						if (topFocus > 0) {
							viewIsCloserToBoundaryV = Math.abs(top - mTempRect.top) <= Math.abs(top - candidateRect.top);
						}
						else {
							viewIsCloserToBoundaryV = Math.abs(bottom - mTempRect.bottom) <= Math.abs(bottom - candidateRect.bottom);
						}
					}
					boolean viewIsCloserToBoundaryH = true;
					if (leftFocus != 0) {
						if (leftFocus > 0) {
							viewIsCloserToBoundaryH = Math.abs(left - mTempRect.left) <= Math.abs(left - candidateRect.left);
						}
						else {
							viewIsCloserToBoundaryH = Math.abs(right - mTempRect.right) <= Math.abs(right - candidateRect.right);
						}
					}
					if (topFocus != 0 && leftFocus != 0) {
						// On privilgie le dplacement dominant
						if (Math.abs(leftFocus) > Math.abs(topFocus)) {
							if ((leftFocus > 0 && mTempRect.left != candidateRect.left) ||
									(leftFocus < 0 && mTempRect.right != candidateRect.right)) {
								viewIsCloserToBoundaryV = viewIsCloserToBoundaryH;
							}
						}
						else if (Math.abs(leftFocus) < Math.abs(topFocus)) {
							if ((topFocus > 0 && mTempRect.top != candidateRect.top) ||
									(topFocus < 0 && mTempRect.bottom != candidateRect.bottom)) {
								viewIsCloserToBoundaryH = viewIsCloserToBoundaryV;
							}
						}
					}
					final boolean viewIsCloserToBoundaries = viewIsCloserToBoundaryV && viewIsCloserToBoundaryH;
					if (foundFullyContainedFocusable) {
						if (viewIsFullyContained && viewIsCloserToBoundaries) {
							/*
							 * We're dealing with only fully contained views, so
							 * it has to be closer to the boundaries to beat our
							 * candidate
							 */
							focusCandidate = view;
							candidateRect.set(mTempRect);
						}
					}
					else {
						if (viewIsFullyContained) {
							/* Any fully contained view beats a partially contained view */
							focusCandidate = view;
							candidateRect.set(mTempRect);
							foundFullyContainedFocusable = true;
						}
						else if (viewIsCloserToBoundaries) {
							/*
							 * Partially contained view beats another partially
							 * contained view if it's closer
							 */
							focusCandidate = view;
							candidateRect.set(mTempRect);
						}
					}
				}
			}
		}

		return focusCandidate;
	}

	/**
	 * <p>Handles scrolling in response to a "page up/down" shortcut press. This
	 * method will scroll the view by one page up or down and give the focus
	 * to the topmost/bottommost component in the new visible area. If no
	 * component is a good candidate for focus, this scrollview reclaims the
	 * focus.</p>
	 *
	 * @param direction the scroll direction: {@link android.view.View#FOCUS_UP}
	 *                  to go one page up or
	 *                  {@link android.view.View#FOCUS_DOWN} to go one page down
	 * @return true if the key event is consumed by this method, false otherwise
	 */
	public boolean pageScroll(int direction) {
		boolean down = direction == View.FOCUS_DOWN;
		int height = getHeight();

		if (down) {
			mTempRect.top = getScrollY() + height;
			int count = getChildCount();
			if (count > 0) {
				View view = getChildAt(count - 1);
				if (mTempRect.top + height > view.getBottom()) {
					mTempRect.top = view.getBottom() - height;
				}
			}
		} else {
			mTempRect.top = getScrollY() - height;
			if (mTempRect.top < 0) {
				mTempRect.top = 0;
			}
		}
		mTempRect.bottom = mTempRect.top + height;

		return scrollAndFocusV(direction, mTempRect.top, mTempRect.bottom);
	}

	/**
	 * <p>Handles scrolling in response to a "home/end" shortcut press. This
	 * method will scroll the view to the top or bottom and give the focus
	 * to the topmost/bottommost component in the new visible area. If no
	 * component is a good candidate for focus, this scrollview reclaims the
	 * focus.</p>
	 *
	 * @param direction the scroll direction: {@link android.view.View#FOCUS_UP}
	 *                  to go the top of the view or
	 *                  {@link android.view.View#FOCUS_DOWN} to go the bottom
	 * @return true if the key event is consumed by this method, false otherwise
	 */
	public boolean fullScroll(int direction) {
		boolean down  = (direction == View.FOCUS_DOWN);
		boolean right = (direction == View.FOCUS_RIGHT);
		int height = getHeight();
		int width = getWidth();

		mTempRect.top = 0;
		mTempRect.bottom = height;
		mTempRect.left = 0;
		mTempRect.right = width;

		if (down) {
			if (getChildCount() > 0) {
				View view = getChildAt(0);
				mTempRect.bottom = view.getBottom();
				mTempRect.top = mTempRect.bottom - height;
			}
		}
		if (right) {
			if (getChildCount() > 0) {
				View view = getChildAt(0);
				mTempRect.right = view.getRight();
				mTempRect.left = mTempRect.right - width;
			}
		}

		if (direction == View.FOCUS_DOWN || direction == View.FOCUS_UP)
			return scrollAndFocusV(direction, mTempRect.top, mTempRect.bottom);
		else
			return scrollAndFocusH(direction, mTempRect.left, mTempRect.right);
	}

	/**
	 * <p>Scrolls the view to make the area defined by <code>top</code> and
	 * <code>bottom</code> visible. This method attempts to give the focus
	 * to a component visible in this area. If no component can be focused in
	 * the new visible area, the focus is reclaimed by this scrollview.</p>
	 *
	 * @param direction the scroll direction: {@link android.view.View#FOCUS_UP}
	 *                  to go upward
	 *                  {@link android.view.View#FOCUS_DOWN} to downward
	 * @param top       the top offset of the new area to be made visible
	 * @param bottom    the bottom offset of the new area to be made visible
	 * @return true if the key event is consumed by this method, false otherwise
	 */
	private boolean scrollAndFocusV(int direction, int top, int bottom) {
		boolean handled = true;

		int height = getHeight();
		int containerTop = getScrollY();
		int containerBottom = containerTop + height;
		boolean goUp = direction == View.FOCUS_UP;

		View newFocused;
		if (goUp)
			newFocused = findFocusableViewInBounds(1, top, bottom, 0, 0, 0);
		else
			newFocused = findFocusableViewInBounds(-1, top, bottom, 0, 0, 0);
		if (newFocused == null) {
			newFocused = this;
		}

		if (top >= containerTop && bottom <= containerBottom) {
			handled = false;
		} else {
			int delta = goUp ? (top - containerTop) : (bottom - containerBottom);
			doScroll(0, delta);
		}

		if (newFocused != findFocus())
			newFocused.requestFocus(direction);

		return handled;
	}

	private boolean scrollAndFocusH(int direction, int left, int right) {
		boolean handled = true;

		int width = getWidth();
		int containerLeft = getScrollX();
		int containerRight = containerLeft + width;
		boolean goLeft = direction == View.FOCUS_LEFT;

		View newFocused;
		if (goLeft)
			newFocused = findFocusableViewInBounds(0, 0, 0, 1, left, right);
		else
			newFocused = findFocusableViewInBounds(0, 0, 0, -1, left, right);
		if (newFocused == null) {
			newFocused = this;
		}

		if (left >= containerLeft && right <= containerRight) {
			handled = false;
		} else {
			int delta = goLeft ? (left - containerLeft) : (right - containerRight);
			doScroll(delta, 0);
		}

		if (newFocused != findFocus())
			newFocused.requestFocus(direction);

		return handled;
	}

	/**
	 * Handle scrolling in response to an up or down arrow click.
	 *
	 * @param direction The direction corresponding to the arrow key that was
	 *                  pressed
	 * @return True if we consumed the event, false otherwise
	 */
	public boolean arrowScrollV(int direction) {
		View currentFocused = findFocus();
		if (currentFocused == this) currentFocused = null;

		View nextFocused = FocusFinder.getInstance().findNextFocus(this, currentFocused, direction);

		final int maxJump = getMaxScrollAmountV();

		if (nextFocused != null && isWithinDeltaOfScreenV(nextFocused, maxJump)) {
			nextFocused.getDrawingRect(mTempRect);
			offsetDescendantRectToMyCoords(nextFocused, mTempRect);
			int scrollDelta = computeScrollDeltaToGetChildRectOnScreenV(mTempRect);
			doScroll(0, scrollDelta);
			nextFocused.requestFocus(direction);
		} else {
			// no new focus
			int scrollDelta = maxJump;

			if (direction == View.FOCUS_UP && getScrollY() < scrollDelta) {
				scrollDelta = getScrollY();
			} else if (direction == View.FOCUS_DOWN) {
				if (getChildCount() > 0) {
					int daBottom = getChildAt(0).getBottom();

					int screenBottom = getScrollY() + getHeight() - getPaddingBottom();

					if (daBottom - screenBottom < maxJump) {
						scrollDelta = daBottom - screenBottom;
					}
				}
			}
			if (scrollDelta == 0) {
				return false;
			}
			doScroll(0, direction == View.FOCUS_DOWN ? scrollDelta : -scrollDelta);
		}

		if (currentFocused != null && currentFocused.isFocused()
				&& isOffScreenV(currentFocused)) {
			// previously focused item still has focus and is off screen, give
			// it up (take it back to ourselves)
			// (also, need to temporarily force FOCUS_BEFORE_DESCENDANTS so we are
			// sure to
			// get it)
			final int descendantFocusability = getDescendantFocusability();  // save
			setDescendantFocusability(ViewGroup.FOCUS_BEFORE_DESCENDANTS);
			requestFocus();
			setDescendantFocusability(descendantFocusability);  // restore
		}
		return true;
	}
	public boolean arrowScrollH(int direction) {
		View currentFocused = findFocus();
		if (currentFocused == this) currentFocused = null;

		View nextFocused = FocusFinder.getInstance().findNextFocus(this, currentFocused, direction);

		final int maxJump = getMaxScrollAmountH();

		if (nextFocused != null && isWithinDeltaOfScreenH(nextFocused, maxJump)) {
			nextFocused.getDrawingRect(mTempRect);
			offsetDescendantRectToMyCoords(nextFocused, mTempRect);
			int scrollDelta = computeScrollDeltaToGetChildRectOnScreenH(mTempRect);
			doScroll(scrollDelta, 0);
			nextFocused.requestFocus(direction);
		} else {
			// no new focus
			int scrollDelta = maxJump;

			if (direction == View.FOCUS_LEFT && getScrollX() < scrollDelta) {
				scrollDelta = getScrollX();
			} else if (direction == View.FOCUS_RIGHT && getChildCount() > 0) {
				int daRight = getChildAt(0).getRight();

				int screenRight = getScrollX() + getWidth() - getPaddingRight();

				if (daRight - screenRight < maxJump) {
					scrollDelta = daRight - screenRight;
				}
			}
			if (scrollDelta == 0) {
				return false;
			}
			doScroll(direction == View.FOCUS_RIGHT ? scrollDelta : -scrollDelta, 0);
		}

		if (currentFocused != null && currentFocused.isFocused()
				&& isOffScreenH(currentFocused)) {
			// previously focused item still has focus and is off screen, give
			// it up (take it back to ourselves)
			// (also, need to temporarily force FOCUS_BEFORE_DESCENDANTS so we are
			// sure to
			// get it)
			final int descendantFocusability = getDescendantFocusability();  // save
			setDescendantFocusability(ViewGroup.FOCUS_BEFORE_DESCENDANTS);
			requestFocus();
			setDescendantFocusability(descendantFocusability);  // restore
		}
		return true;
	}

	/**
	 * @return whether the descendant of this scroll view is scrolled off
	 *  screen.
	 */
	private boolean isOffScreenV(View descendant) {
		return !isWithinDeltaOfScreenV(descendant, 0);
	}
	private boolean isOffScreenH(View descendant) {
		return !isWithinDeltaOfScreenH(descendant, 0);
	}

	/**
	 * @return whether the descendant of this scroll view is within delta
	 *  pixels of being on the screen.
	 */
	private boolean isWithinDeltaOfScreenV(View descendant, int delta) {
		descendant.getDrawingRect(mTempRect);
		offsetDescendantRectToMyCoords(descendant, mTempRect);

		return (mTempRect.bottom + delta) >= getScrollY()
				&& (mTempRect.top - delta) <= (getScrollY() + getHeight());
	}
	private boolean isWithinDeltaOfScreenH(View descendant, int delta) {
		descendant.getDrawingRect(mTempRect);
		offsetDescendantRectToMyCoords(descendant, mTempRect);

		return (mTempRect.right + delta) >= getScrollX()
				&& (mTempRect.left - delta) <= (getScrollX() + getWidth());
	}

	/**
	 * Smooth scroll by a Y delta
	 *
	 * @param delta the number of pixels to scroll by on the Y axis
	 */
	private void doScroll(int deltaX, int deltaY) {
		if (deltaX != 0 || deltaY != 0) {
			if (mSmoothScrollingEnabled) {
				smoothScrollBy(deltaX, deltaY);
			} else {
				scrollBy(deltaX, deltaY);
			}
		}
	}

	/**
	 * Like {@link View#scrollBy}, but scroll smoothly instead of immediately.
	 *
	 * @param dx the number of pixels to scroll by on the X axis
	 * @param dy the number of pixels to scroll by on the Y axis
	 */
	public final void smoothScrollBy(int dx, int dy) {
		if (getChildCount() == 0) {
			// Nothing to do.
			return;
		}
		long duration = AnimationUtils.currentAnimationTimeMillis() - mLastScroll;
		if (duration > ANIMATED_SCROLL_GAP) {
			final int width = getWidth() - getPaddingRight() - getPaddingLeft();
			final int right = getChildAt(0).getWidth();
			final int maxX = Math.max(0, right - width);
			final int scrollX = getScrollX();
			dx = Math.max(0, Math.min(scrollX + dx, maxX)) - scrollX; 

			final int height = getHeight() - getPaddingBottom() - getPaddingTop();
			final int bottom = getChildAt(0).getHeight();
			final int maxY = Math.max(0, bottom - height);
			final int scrollY = getScrollY();
			dy = Math.max(0, Math.min(scrollY + dy, maxY)) - scrollY;

			mScroller.startScroll(scrollX, scrollY, dx, dy);
			postInvalidate();
		} else {
			if (!mScroller.isFinished()) {
				mScroller.abortAnimation();
			}
			scrollBy(dx, dy);
		}
		mLastScroll = AnimationUtils.currentAnimationTimeMillis();
	}

	/**
	 * Like {@link #scrollTo}, but scroll smoothly instead of immediately.
	 *
	 * @param x the position where to scroll on the X axis
	 * @param y the position where to scroll on the Y axis
	 */
	public final void smoothScrollTo(int x, int y) {
		smoothScrollBy(x - getScrollX(), y - getScrollY());
	}

	/**
	 * <p>The scroll range of a scroll view is the overall height of all of its
	 * children.</p>
	 */
	@Override
	protected int computeHorizontalScrollRange() {
		final int count = getChildCount();
		final int contentWidth = getWidth() - getPaddingLeft() - getPaddingRight();
		if (count == 0) {
			return contentWidth;
		}

		int scrollRange = getChildAt(0).getRight();
		final int scrollX = getScrollX();
		final int overscrollRight = Math.max(0, scrollRange - contentWidth);
		if (scrollX < 0) {
			scrollRange -= scrollX;
		} else if (scrollX > overscrollRight) {
			scrollRange += scrollX - overscrollRight;
		} 

		return scrollRange;
	}
	@Override    
	protected int computeVerticalScrollRange() {
		final int count = getChildCount();
		final int contentHeight = getHeight() - getPaddingBottom() - getPaddingTop();
		if (count == 0) {
			return contentHeight;
		}

		int scrollRange = getChildAt(0).getBottom();
		final int scrollY = getScrollY();
		final int overscrollBottom = Math.max(0, scrollRange - contentHeight);
		if (scrollY < 0) {
			scrollRange -= scrollY;
		} else if (scrollY > overscrollBottom) {
			scrollRange += scrollY - overscrollBottom;
		}

		return scrollRange;
	}

	@Override
	protected int computeHorizontalScrollOffset() {
		return Math.max(0, super.computeHorizontalScrollOffset());
	}
	@Override
	protected int computeVerticalScrollOffset() {
		return Math.max(0, super.computeVerticalScrollOffset());
	}

	@Override
	protected void measureChild(View child, int parentWidthMeasureSpec, int parentHeightMeasureSpec) {
		ViewGroup.LayoutParams lp = child.getLayoutParams();

		int childWidthMeasureSpec;
		int childHeightMeasureSpec;

		childWidthMeasureSpec = getChildMeasureSpec(parentWidthMeasureSpec, getPaddingLeft()
				+ getPaddingRight(), lp.width);

		childHeightMeasureSpec = getChildMeasureSpec(parentHeightMeasureSpec, getPaddingTop()
				+ getPaddingBottom(), lp.height);

		child.measure(childWidthMeasureSpec, childHeightMeasureSpec);
	}

	@Override
	protected void measureChildWithMargins(View child, int parentWidthMeasureSpec, int widthUsed,
			int parentHeightMeasureSpec, int heightUsed) {
		final MarginLayoutParams lp = (MarginLayoutParams) child.getLayoutParams();

		final int childWidthMeasureSpec = getChildMeasureSpec(parentWidthMeasureSpec,
				getPaddingLeft() + getPaddingRight() + lp.leftMargin + lp.rightMargin
				+ widthUsed, lp.width);
		final int childHeightMeasureSpec = getChildMeasureSpec(parentHeightMeasureSpec,
				getPaddingTop() + getPaddingBottom() + lp.topMargin + lp.bottomMargin
				+ heightUsed, lp.height);

		child.measure(childWidthMeasureSpec, childHeightMeasureSpec);
	}

	@Override
	public void computeScroll() {
		if (mScroller.computeScrollOffset()) {
			// This is called at drawing time by ViewGroup.  We don't want to
			// re-show the scrollbars at this point, which scrollTo will do,
			// so we replicate most of scrollTo here.
			//
			//         It's a little odd to call onScrollChanged from inside the drawing.
			//
			//         It is, except when you remember that computeScroll() is used to
			//         animate scrolling. So unless we want to defer the onScrollChanged()
			//         until the end of the animated scrolling, we don't really have a
			//         choice here.
			//
			//         I agree.  The alternative, which I think would be worse, is to post
			//         something and tell the subclasses later.  This is bad because there
			//         will be a window where getScrollX()/Y is different from what the app
			//         thinks it is.
			//
			int oldX = getScrollX();
			int oldY = getScrollY();
			int x = mScroller.getCurrX();
			int y = mScroller.getCurrY();

			if (getChildCount() > 0) {
				View child = getChildAt(0);
				x = clamp(x, getWidth() - getPaddingRight() - getPaddingLeft(), child.getWidth());
				y = clamp(y, getHeight() - getPaddingBottom() - getPaddingTop(), child.getHeight());
				if (x != oldX || y != oldY) {
					super.scrollTo(x, y);
				}
			}
			if (!awakenScrollBars()) {
				// Keep on drawing until the animation has finished.
				postInvalidate();
			}
		}
	}

	/**
	 * Scrolls the view to the given child.
	 *
	 * @param child the View to scroll to
	 */
	private void scrollToChild(View child) {
		child.getDrawingRect(mTempRect);

		/* Offset from child's local coordinates to ScrollView coordinates */
		offsetDescendantRectToMyCoords(child, mTempRect);

		int scrollDeltaX = computeScrollDeltaToGetChildRectOnScreenH(mTempRect);
		int scrollDeltaY = computeScrollDeltaToGetChildRectOnScreenV(mTempRect);

		if (scrollDeltaX != 0 || scrollDeltaY != 0) {
			scrollBy(scrollDeltaX, scrollDeltaY);
		}
	}

	/**
	 * If rect is off screen, scroll just enough to get it (or at least the
	 * first screen size chunk of it) on screen.
	 *
	 * @param rect      The rectangle.
	 * @param immediate True to scroll immediately without animation
	 * @return true if scrolling was performed
	 */
	private boolean scrollToChildRect(Rect rect, boolean immediate) {
		final int deltaX = computeScrollDeltaToGetChildRectOnScreenH(rect);
		final int deltaY = computeScrollDeltaToGetChildRectOnScreenV(rect);
		final boolean scroll = (deltaX != 0 || deltaY != 0);
		if (scroll) {
			if (immediate) {
				scrollBy(deltaX, deltaY);
			} else {
				smoothScrollBy(deltaX, deltaY);
			}
		}
		return scroll;
	}

	/**
	 * Compute the amount to scroll in the Y direction in order to get
	 * a rectangle completely on the screen (or, if taller than the screen,
	 * at least the first screen size chunk of it).
	 *
	 * @param rect The rect.
	 * @return The scroll delta.
	 */
	protected int computeScrollDeltaToGetChildRectOnScreenV(Rect rect) {
		if (getChildCount() == 0) return 0;

		int height = getHeight();
		int screenTop = getScrollY();
		int screenBottom = screenTop + height;

		int fadingEdge = getVerticalFadingEdgeLength();

		// leave room for top fading edge as long as rect isn't at very top
		if (rect.top > 0) {
			screenTop += fadingEdge;
		}

		// leave room for bottom fading edge as long as rect isn't at very bottom
		if (rect.bottom < getChildAt(0).getHeight()) {
			screenBottom -= fadingEdge;
		}

		int scrollYDelta = 0;

		if (rect.bottom > screenBottom && rect.top > screenTop) {
			// need to move down to get it in view: move down just enough so
			// that the entire rectangle is in view (or at least the first
			// screen size chunk).

			if (rect.height() > height) {
				// just enough to get screen size chunk on
				scrollYDelta += (rect.top - screenTop);
			} else {
				// get entire rect at bottom of screen
				scrollYDelta += (rect.bottom - screenBottom);
			}

			// make sure we aren't scrolling beyond the end of our content
			int bottom = getChildAt(0).getBottom();
			int distanceToBottom = bottom - screenBottom;
			scrollYDelta = Math.min(scrollYDelta, distanceToBottom);

		} else if (rect.top < screenTop && rect.bottom < screenBottom) {
			// need to move up to get it in view: move up just enough so that
			// entire rectangle is in view (or at least the first screen
			// size chunk of it).

			if (rect.height() > height) {
				// screen size chunk
				scrollYDelta -= (screenBottom - rect.bottom);
			} else {
				// entire rect at top
				scrollYDelta -= (screenTop - rect.top);
			}

			// make sure we aren't scrolling any further than the top our content
			scrollYDelta = Math.max(scrollYDelta, -getScrollY());
		}
		return scrollYDelta;
	}
	protected int computeScrollDeltaToGetChildRectOnScreenH(Rect rect) {
		if (getChildCount() == 0) return 0;

		int width = getWidth();
		int screenLeft = getScrollX();
		int screenRight = screenLeft + width;

		int fadingEdge = getHorizontalFadingEdgeLength();

		// leave room for left fading edge as long as rect isn't at very left
		if (rect.left > 0) {
			screenLeft += fadingEdge;
		}

		// leave room for right fading edge as long as rect isn't at very right
		if (rect.right < getChildAt(0).getWidth()) {
			screenRight -= fadingEdge;
		}

		int scrollXDelta = 0;

		if (rect.right > screenRight && rect.left > screenLeft) {
			// need to move right to get it in view: move right just enough so
			// that the entire rectangle is in view (or at least the first
			// screen size chunk).

			if (rect.width() > width) {
				// just enough to get screen size chunk on
				scrollXDelta += (rect.left - screenLeft);
			} else {
				// get entire rect at right of screen
				scrollXDelta += (rect.right - screenRight);
			}

			// make sure we aren't scrolling beyond the end of our content
			int right = getChildAt(0).getRight();
			int distanceToRight = right - screenRight;
			scrollXDelta = Math.min(scrollXDelta, distanceToRight);

		} else if (rect.left < screenLeft && rect.right < screenRight) {
			// need to move right to get it in view: move right just enough so that
			// entire rectangle is in view (or at least the first screen
			// size chunk of it).

			if (rect.width() > width) {
				// screen size chunk
				scrollXDelta -= (screenRight - rect.right);
			} else {
				// entire rect at left
				scrollXDelta -= (screenLeft - rect.left);
			}

			// make sure we aren't scrolling any further than the left our content
			scrollXDelta = Math.max(scrollXDelta, -getScrollX());
		}
		return scrollXDelta;
	}

	@Override
	public void requestChildFocus(View child, View focused) {
		if (!mIsLayoutDirty) {
			scrollToChild(focused);
		} else {
			// The child may not be laid out yet, we can't compute the scroll yet
			mChildToScrollTo = focused;
		}
		super.requestChildFocus(child, focused);
	}

	/**
	 * When looking for focus in children of a scroll view, need to be a little
	 * more careful not to give focus to something that is scrolled off screen.
	 *
	 * This is more expensive than the default {@link android.view.ViewGroup}
	 * implementation, otherwise this behavior might have been made the default.
	 */
	@Override
	protected boolean onRequestFocusInDescendants(int direction, Rect previouslyFocusedRect) {
		//		// convert from forward / backward notation to up / down / left / right
		//		// (ugh).
		//		if (direction == View.FOCUS_FORWARD) {
		//			direction = View.FOCUS_DOWN;
		//		} else if (direction == View.FOCUS_BACKWARD) {
		//			direction = View.FOCUS_UP;
		//		}
		//
		//		final View nextFocus = previouslyFocusedRect == null ?
		//				FocusFinder.getInstance().findNextFocus(this, null, direction) :
		//					FocusFinder.getInstance().findNextFocusFromRect(this, previouslyFocusedRect, direction);
		//
		//		if (nextFocus == null) {
		//			return false;
		//		}
		//
		//		if (isOffScreen(nextFocus)) {
		//			return false;
		//		}
		//
		//		return nextFocus.requestFocus(direction, previouslyFocusedRect);
		return false;
	}    

	@Override
	public boolean requestChildRectangleOnScreen(View child, Rect rectangle, boolean immediate) {
		// offset into coordinate space of this scroll view
		rectangle.offset(child.getLeft() - child.getScrollX(),
				child.getTop() - child.getScrollY());

		return scrollToChildRect(rectangle, immediate);
	}

	@Override
	public void requestLayout() {
		mIsLayoutDirty = true;
		super.requestLayout();
	}

	@Override
	protected void onLayout(boolean changed, int l, int t, int r, int b) {
		super.onLayout(changed, l, t, r, b);
		mIsLayoutDirty = false;
		// Give a child focus if it needs it 
		if (mChildToScrollTo != null && isViewDescendantOf(mChildToScrollTo, this)) {
			scrollToChild(mChildToScrollTo);
		}
		mChildToScrollTo = null;

		// Calling this with the present values causes it to re-claim them
		scrollTo(getScrollX(), getScrollY());
	}

	@Override
	protected void onSizeChanged(int w, int h, int oldw, int oldh) {
		super.onSizeChanged(w, h, oldw, oldh);

		View currentFocused = findFocus();
		if (currentFocused == null || currentFocused == this)
			return;

		// If the currently-focused view was visible on the screen when the
		// screen was at the old height, then scroll the screen to make that
		// view visible with the new screen height.
		currentFocused.getDrawingRect(mTempRect);
		offsetDescendantRectToMyCoords(currentFocused, mTempRect);
		int scrollDeltaX = computeScrollDeltaToGetChildRectOnScreenH(mTempRect);
		int scrollDeltaY = computeScrollDeltaToGetChildRectOnScreenV(mTempRect);
		doScroll(scrollDeltaX, scrollDeltaY);
	}    

	/**
	 * Return true if child is a descendant of parent, (or equal to the parent).
	 */
	private static boolean isViewDescendantOf(View child, View parent) {
		if (child == parent) {
			return true;
		}

		final ViewParent theParent = child.getParent();
		return (theParent instanceof ViewGroup) && isViewDescendantOf((View) theParent, parent);
	}    

	/**
	 * Fling the scroll view
	 *
	 * @param velocityY The initial velocity in the Y direction. Positive
	 *                  numbers mean that the finger/cursor is moving down the screen,
	 *                  which means we want to scroll towards the top.
	 */
	public void fling(int velocityX, int velocityY) {
		if (getChildCount() > 0) {
			int width = getWidth() - getPaddingRight() - getPaddingLeft();
			int right = getChildAt(0).getWidth();
			int height = getHeight() - getPaddingBottom() - getPaddingTop();
			int bottom = getChildAt(0).getHeight();

			final int oldX = getScrollX();
			final int oldY = getScrollY();

			mScroller.fling(oldX, oldY, velocityX, velocityY, 0, Math.max(0, right - width), 0, Math.max(0, bottom - height));

			FindNewFocusedView(mScroller.getFinalX() - oldX, mScroller.getFinalY() - oldY);

			postInvalidate();
		}
	}

	protected boolean ScrollingIsFinished()
	{
		return mScroller.isFinished();
	}

	/**
	 * {@inheritDoc}
	 *
	 * <p>This version also clamps the scrolling to the bounds of our child.
	 */
	@Override
	public void scrollTo(int x, int y) {
		// we rely on the fact the View.scrollBy calls scrollTo.
		if (getChildCount() > 0) {
			View child = getChildAt(0);
			x = clamp(x, getWidth() - getPaddingRight() - getPaddingLeft(), child.getWidth());
			y = clamp(y, getHeight() - getPaddingBottom() - getPaddingTop(), child.getHeight());
			if (x != getScrollX() || y != getScrollY()) {
				super.scrollTo(x, y);
			}
		}
	}

	private int clamp(int n, int my, int child) {
		if (my >= child || n < 0) {
			/* my >= child is this case:
			 *                    |--------------- me ---------------|
			 *     |------ child ------|
			 * or
			 *     |--------------- me ---------------|
			 *            |------ child ------|
			 * or
			 *     |--------------- me ---------------|
			 *                                  |------ child ------|
			 *
			 * n < 0 is this case:
			 *     |------ me ------|
			 *                    |-------- child --------|
			 *     |-- mScrollX --|
			 */
			return 0;
		}
		if ((my+n) > child) {
			/* this case:
			 *                    |------ me ------|
			 *     |------ child ------|
			 *     |-- mScrollX --|
			 */
			return child-my;
		}
		return n;
	}
}